if (randomizeInitValues)
{
	global beta = Random (0.1,10);	
}
else
{
	global beta = .25;
}
categDef2 = " ";

if (USE_CONST_ON_PART)
{
beta:>0.01;beta:<100;
category d =  (resp2, EQUAL, MEAN, 
				GammaDist(_x_,beta,beta), 
				CGammaDist(_x_,beta,beta), 
				0 , 
				1e25,
				CGammaDist(_x_,beta+1,beta),
				CONSTANT_ON_PARTITION);
}
else
{
beta:>0.01;beta:<100;
category d =  (resp2, EQUAL, MEAN, 
				GammaDist(_x_,beta,beta), 
				CGammaDist(_x_,beta,beta), 
				0 , 
				1e25,
				CGammaDist(_x_,beta+1,beta));

}
